/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/****************************************************************************
*                                                                           *
* CONE                                                                      *
*                                                                           *
* Control Flow Event Notification Engine                                    *
*                                                                           *
****************************************************************************/

#include <cstdlib>    // for atoi() call
#include <dpcl.h>
#include <fstream>
   using std::ofstream;
#include <iostream>
   using std::cerr;
   using std::cout;
#include <string>
   using std::string;
#include <libgen.h>   // for basename() call


#include "Executable.h"
#include "Callgraph.h"
#include "cone_error.h"
#include "cone_probes.h"

#define MAX_STR_LEN 4096 
/****************************************************************************
* global variables
****************************************************************************/

PoeAppl pappl;
Process proc;

ProbeExp init_pfun;
ProbeExp start_pfun;
ProbeExp term_pfun;
ProbeExp enter_pfun;
ProbeExp exit_pfun;
ProbeExp before_pfun;
ProbeExp after_pfun;
ProbeExp mpi_init_pfun;
ProbeExp mpi_finalize_pfun;

string mainfun = "main";

Executable* exec = NULL;

Callgraph* cgraph = NULL;
                       
// poe application ?
bool ispoe  = true;

// default poe name
char* poe_name = "/usr/bin/poe";

// default linking option 
bool isdynamic = true ;   

// default host name
char* host_name = "localhost";

// probe_module serial applications
char* probe_module = "libcone.so" ; 

// probe_module mpi applications
char* probe_module_mpi = "libcone.mpi.so" ; 

// default program name
char*  prog_name = "a.out";

int nprocs = 0; 
int numevs = 0;

// default CONE_METRICS environment variable
char* cone_metrics = "CONE_METRICS=CYCLES:L1_D_MISS:L1_D_ACCESS:TLB_MISS:TLB_I_MISS:TLB_D_MISS:INSTRUCTION" ;

// argument counter
int a = 1; 

// status
AisStatus sts;


/*************************************************************************** 
             termination of processes
****************************************************************************/

void terminate_application()
{
  if ( ispoe )
    pappl.bdestroy();
  else
    proc.bdestroy();
}

/****************************************************************************
* call-back functions
****************************************************************************/

void
data_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  cout << (char*) msg; 
}

void
stdout_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  cout << (char*) msg; 
}

void
stderr_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  cerr << (char*) msg; 
}

void
exit_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  if ( ispoe )
    {
      nprocs--;
      if ( nprocs == 0 )
        Ais_end_main_loop();
    }
  else
    Ais_end_main_loop(); 
}

int sigint_handler(int s)
{
  cerr <<"\n CONE_DPCL terminated.\n" ;
  terminate_application();
  return 0;
}

/****************************************************************************
* parse command line 
****************************************************************************/

void parse_cmdl(int argc, char *argv[])
{
  bool error = false;
  
  while ( a < argc && argv[a][0] == '-' ) 
    {
      // is mpi application?
      if ( strcmp(argv[a], "-poe") == 0 )
	{
	  ispoe = true;
	}
      // is serial application?
      else if ( strcmp(argv[a], "-nopoe") == 0 )
	{
	  ispoe = false;
	}
      // numer of processes
      else if ( strcmp(argv[a], "-procs") == 0 ) 
	{
	  ispoe = true;

	  if ( (a+1) < argc )
	    { 
	      nprocs = atoi(argv[++a]); 
	    }
	  else
	    {
	      cone_warning("missing value for option -procs");
	      error = true;
	    }
	  
	  if ( nprocs < 1 )
	    {
	      cone_warning("invalid value for option -procs");
	      error = true;
	    }
	}
      // name of main function
      else if ( strcmp(argv[a], "-main") == 0 ) 
	{
	  if ( (a+1) < argc ) 
	    {
	      mainfun = argv[++a];
	    }
	  else
	    {
	      cone_warning("missing value for option -main");
	      error = true;
	    }
	}
      // linking options 
      else if ( strcmp(argv[a], "-static") == 0 )
        {
          isdynamic = false ; 
        }     
      else if ( strcmp(argv[a], "-dynamic") == 0 )
	{
          if ( (a+1) < argc )
            {   
              probe_module = argv[++a]; 
            } 
          else
	    {
	      cone_warning("missing value for option -dynamic");
	      error = true;
	    }
        } 
      // event set option
      else if ( strcmp(argv[a], "-evs") == 0  )
        {
          if ( (a+1) < argc )
	    { 
	      numevs = atoi(argv[++a]); 
	    }
	  else
	    {
	      cone_warning("missing value for option -evs");
	      error = true;
	    }
	  
	  if ( numevs < 0 )
	    {
	      cone_warning("invalid value for option -evs");
	      error = true;
	    }      
          else if ( numevs > 7 )
            {
              cone_warning("invalid value for option -evs");
	      error = true;
            } 
          else if ( numevs == 0 ) 
	    {
              cone_metrics = "CONE_METRICS=CYCLES:L1_D_ACCESS:L1_D_MISS:TLB_MISS:TLB_I_MISS:TLB_D_MISS:INSTRUCTION" ;  
	    }
	  else if ( numevs == 1 )
	    {
              cone_metrics = "CONE_METRICS=CYCLES:INSTRUCTION:L1_D_READ:L1_D_WRITE" ; 
	    }
	  else if ( numevs == 2 )
	    {
              cone_metrics = "CONE_METRICS=CYCLES:INSTRUCTION:INTEGER" ; 
	    }
	  else if ( numevs == 3 )
	    {
              cone_metrics = "CONE_METRICS=CYCLES:INSTRUCTION:FP_INS:FP_SQRT_INS:FP_DIV_INS";
	    } 	  
	  else if ( numevs == 4 ) 
	    {
	      cone_metrics = "CONE_METRICS=INSTRUCTION:FP_INS";
	    }	  
	  else if ( numevs == 5 ) 
	    {
	      cone_metrics = "CONE_METRICS=L1_D_ACCESS:L1_D_MISS";
	    }
	  else if ( numevs == 6 ) 
	    {
	      cone_metrics = "CONE_METRICS=L1_D_MISS";
	    }
	  else if ( numevs == 7 ) 
	    {
	      cone_metrics = "CONE_METRICS=FP_INS";
	    }
        }
      // host name
      else if ( strcmp(argv[a], "-host") == 0 ) 
	{
	  if ( (a+1) < argc ) 
	    {
	      host_name = argv[++a];
	    }
	  else
	    {
	      cone_warning("missing value for option -host");
	      error = true;
	    }
 	}
      else if ( strcmp(argv[a], "-help") == 0 ) 
	{
	  error = true;
	  break;
	}
      ++a;
    }
  
  if ( a >= argc )
    {
      cone_warning("missing executable");
      error = true;
    }
  else
    {
      prog_name = argv[a++];
      putenv(cone_metrics);
    }

  if ( error )
    {
      cerr << "Usage: " << argv[0] << " [-static|-dynamic <filename>]  [-poe|-nopoe] [-procs <n>]" << endl 
           << "       [-evs <n>] " << endl
     	   << "       [-host <name>] [-main <name>] <executable>" << endl ;
      cerr << "   or: " << argv[0] << " [-help]" << endl ;
      exit(1);
    }
}

/****************************************************************************
* inititalize DPCL
****************************************************************************/

void init_dpcl()
{
  cerr << "Initializing DPCL...            ";

  Ais_initialize();
  
  Ais_add_signal(SIGINT, sigint_handler);
  
  sts = Ais_override_default_callback(AIS_PROC_TERMINATE_MSG, &exit_cb, NULL, NULL, NULL);
  if (sts.status() != ASC_success) 
    cone_error_msg("call-back override failed: %s", sts.status_name());
  

  cerr << "done.\n";
}

/******************************************************************************
 * create application
 *****************************************************************************/

void create_application(int argc, char *argv[])
{
  cerr << "Creating application...         "; 

  
  // arguments
  vector<char*> args(argc - a  + 3);
  if ( ispoe )
    {
      args[0] = poe_name;
      args[1] = prog_name;   
      for ( int i = 0; i < argc - a ; i++ ) 
        args[i + 2] = argv[i + a];
      args[argc - a + 2] = NULL;
    }
  else
    {
      args[0] = prog_name;
      for ( int i = 0; i < argc - a ; i++ ) 
        args[i + 1] = argv[i + a];
      args[argc - a + 2] = NULL;
    }

 
  if ( ispoe )
    {
      // set environment variable M_PROCS if -procs has been
      // specified
      
      if ( nprocs != 0 )
        {
          char tmp[1024];
          sprintf(tmp, "MP_PROCS=%d", nprocs);
          putenv(tmp);
	}


      sts = pappl.bcreate(host_name, 
			  poe_name,
			  &args[0],
			  environ, 
			  stdout_cb,
			  (GCBTagType) 0,
			  stderr_cb,
			  (GCBTagType) 0		    
			  );

      // if number of processes wasn't specified using -procs option,
      // nprocs is assigned the correct value here
      nprocs = pappl.get_count();

      if (sts.status() != ASC_success) 
        {
          terminate_application();  
	  cone_error_msg("PoeAppl::bcreate failed: %s", sts.status_name());
        }
      
      proc = pappl.get_process(0);
    }
  else
    {
       sts = proc.bcreate(host_name, 
			  prog_name,
			  &args[0], 
			  environ, 
			  stdout_cb,
			  (GCBTagType) 0,
			  stderr_cb,
			  (GCBTagType) 0		    
			  );
      if (sts.status() != ASC_success)
        { 
          terminate_application();
	  cone_error_msg("Process::bcreate failed: %s", sts.status_name());
        }
    }

  cerr << "done.\n";
}

/****************************************************************************
* locate probe functions
****************************************************************************/

void locate_probe_functions()
{
  cerr << "Locating probe functions...     ";

  if ( !isdynamic )
     {   
       SourceObj prog = proc.get_program_object();
       
       for (int c = 0 ; c < prog.child_count(); c++)
	 {
	   SourceObj modobj = prog.child(c);
	   
	   vector<char> mod_name(modobj.module_name_length());
	   modobj.module_name(&mod_name[0], mod_name.size());
	   
	   if ( strcmp(&mod_name[0], "cone_probes.c") == 0 )
	     {
	       int npfuns = 0; 
      
	       for ( int d = 0; d < modobj.child_count(); d++) 
		 {
		   SourceObj funcobj = modobj.child(d);
		   
		   vector<char> func_name(funcobj.get_demangled_name_length());
		   funcobj.get_demangled_name(&func_name[0], func_name.size());
		   
		   if ( strcmp(&func_name[0], "cone_init" ) == 0 ) 
		     init_pfun = funcobj.reference();
		   else if ( strcmp(&func_name[0], "cone_start" ) == 0 ) 
	             start_pfun = funcobj.reference();
		   else if ( strcmp(&func_name[0], "cone_enter" ) == 0 )
		     enter_pfun = funcobj.reference();
		   else if ( strcmp(&func_name[0], "cone_exit" ) == 0 )
		     exit_pfun = funcobj.reference();
		   else if ( strcmp(&func_name[0], "cone_before" ) == 0 )
		     before_pfun = funcobj.reference();
		   else if ( strcmp(&func_name[0], "cone_after" ) == 0  ) 
		     after_pfun = funcobj.reference();
		   else if ( strcmp(&func_name[0], "cone_after_mpi_init" ) == 0  ) 
		     mpi_init_pfun = funcobj.reference();
		   else if ( strcmp(&func_name[0], "cone_before_mpi_finalize" ) == 0  ) 
		     mpi_finalize_pfun = funcobj.reference();
		   else
		     continue;
		   
		   npfuns++;
		 }
	       // verify that all probe functions have been found
	       if ( npfuns < 8 )       
		 {
		   terminate_application();
		   cone_error_msg("missing probe function");
		 }
	     }  
	 }
     }
  else if ( isdynamic )
    { 
      int   npfuns = 0;        
      char  tmp[MAX_STR_LEN];
      int   len; 
      char* env ;   
 
      AisStatus sts; 
      char* probe_mod_name;

     
      if ( ispoe )
	probe_mod_name = probe_module_mpi;
      else
	probe_mod_name = probe_module;
 	
      ProbeModule my_probe_mod;

	
      len = strlen(CONE_PM_PATH); 

      if ( CONE_PM_PATH[len-1] == '/' )
        sprintf(tmp, "%s%s", CONE_PM_PATH, probe_mod_name);
      else
        sprintf(tmp, "%s/%s", CONE_PM_PATH, probe_mod_name);  

      try {
	my_probe_mod = ProbeModule(tmp);
      } catch(AisStatus sts) {
	terminate_application();
	cone_error_msg("cannot create probe module: %s", sts.status_name());          
      } 
	
      if ( ispoe )
        sts = pappl.bload_module(&my_probe_mod);
      else
        sts = proc.bload_module(&my_probe_mod);
 
      if (sts.status() != ASC_success)
        {	  
	  terminate_application();
	  cone_error_msg("bload failed: %s", sts.status_name());    
        } 

       for (int i=0 ; i < my_probe_mod.get_count() ; i++)
	 {
	   int size = my_probe_mod.get_name_length(i);
	   vector<char> name(size); 
	   char *funct_name = my_probe_mod.get_name(i , &name[0] , size);
	   	   
	   if ( strcmp(funct_name, "cone_init") == 0 )
	     init_pfun = my_probe_mod.get_reference(i);
	   else if ( strcmp(funct_name , "cone_start") == 0 )
	     start_pfun = my_probe_mod.get_reference(i);
	   else if ( strcmp(funct_name , "cone_enter") == 0 )
	     enter_pfun = my_probe_mod.get_reference(i);
	   else if ( strcmp(funct_name , "cone_exit") == 0 )
	     exit_pfun = my_probe_mod.get_reference(i);
	   else if ( strcmp(funct_name , "cone_before") == 0 )
	     before_pfun = my_probe_mod.get_reference(i);
	   else if ( strcmp(funct_name , "cone_after") == 0 )
	     after_pfun = my_probe_mod.get_reference(i);
	   else if ( strcmp(funct_name, "cone_after_mpi_init" ) == 0 ) 
	     mpi_init_pfun = my_probe_mod.get_reference(i);
	   else if ( strcmp(funct_name, "cone_before_mpi_finalize" ) == 0 ) 
	     mpi_finalize_pfun = my_probe_mod.get_reference(i);
	   else
	     continue ;     
	   
	   npfuns++ ;
	 }         
       if ( npfuns < 8 )
	 {
	   terminate_application();
	   cone_error_msg("missing probe function");
	 }
    }
  cerr << "done.\n";
}

/****************************************************************************
* install & activate probe
****************************************************************************/

void install_probe(ProbeExp& pexp, InstPoint& point)
{
  AisStatus sts;
  GCBTagType tag = (GCBTagType) 0;
  ProbeHandle phandle;
  GCBFuncType cb[1] = { data_cb };

  if ( ispoe )
    {
      sts = pappl.binstall_probe(1, &pexp, &point, 0, 0, &phandle);
      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  cone_error_msg("PoeAppl::binstall_probe failed: %s", sts.status_name());
	}
      sts = pappl.bactivate_probe(1, &phandle);
      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  cone_error_msg("PoeAppl::bactivate_probe failed: %s", sts.status_name());
	}
    }
  else
    {
      sts = proc.binstall_probe(1, &pexp, &point, cb, &tag, &phandle);
      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  cone_error_msg("Process::binstall_probe failed: %s", sts.status_name());
	}
      sts = proc.bactivate_probe(1, &phandle);
      if (sts.status() != ASC_success)
	{
	  terminate_application(); 
	  cone_error_msg("Process::bactivate_probe failed: %s", sts.status_name());
	}
    }
}

/****************************************************************************
* instrument user functions
****************************************************************************/

void instrument_functions()
{
  cerr << "Instrumenting user functions... ";
   
  for ( int i = 0; i < exec->get_num_funcs(); i++ )
    {
      Function* func = exec->get_func(i);
    
      ProbeExp exp(func->id);
      if ( !func->isuser )
	continue;

      if ( func->name  == mainfun )
	{
	  ProbeExp start_call = start_pfun.call(1, &exp);
      	  install_probe(start_call, func->entry);
        }

      if ( func->isuser )
        {  
          ProbeExp exit_call  = exit_pfun.call(1, &exp);	  
          install_probe(exit_call, func->exit);	
        }            
    }
  
  cerr << "done.\n";
}

/****************************************************************************
* instrument call sites
****************************************************************************/

void instrument_call_sites()
{
  cerr << "Instrumenting call sites...     " ;

  for ( int i = 1; i < exec->get_num_csites(); i++ )
    {
      Callsite* csite = exec->get_csite(i);
      
      ProbeExp exp(csite->id);

      if ( !csite->isuser )
	continue;
      
      if (csite->callee->is_mpi_init() )
        {
          ProbeExp mpi_init_call = mpi_init_pfun.call(1, &exp);
          install_probe(mpi_init_call, csite->after);                    

          ProbeExp before_call = before_pfun.call(1, &exp);
	  ProbeExp after_call  = after_pfun.call(1, &exp);

	  install_probe(before_call, csite->before);
          install_probe(after_call, csite->after);	
        }
      else if (csite->callee->is_mpi_finalize() )
        { 
          ProbeExp mpi_finalize_call = mpi_finalize_pfun.call(1, &exp);
          install_probe(mpi_finalize_call, csite->before);
	  
          ProbeExp before_call = before_pfun.call(1, &exp);
	  ProbeExp after_call  = after_pfun.call(1, &exp);

	  install_probe(before_call, csite->before);
          install_probe(after_call, csite->after);	
        } 
      else if ( !csite->callee->isuser && !csite->callee->is_mpi_init() ||
                !csite->callee->isuser && !csite->callee->is_mpi_finalize()  )
	{ 
  	  ProbeExp before_call = before_pfun.call(1, &exp);
	  ProbeExp after_call  = after_pfun.call(1, &exp);

	  install_probe(before_call, csite->before);
          install_probe(after_call, csite->after);		     
	}
      else 
	{ 
	  ProbeExp before_call = before_pfun.call(1, &exp);
          install_probe(before_call, csite->before);	  
	}
    }
  cerr << "done.\n";
}

/****************************************************************************
* initialize probe module
****************************************************************************/

void init_probes()
{
  cerr << "Initializing probe module...    ";

  AisStatus sts;

  ProbeExp init_call = init_pfun.call(0, NULL);

  if ( ispoe )
    {
      sts = pappl.bexecute(init_call, &data_cb, (GCBTagType) 0);
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  cone_error_msg("PoeAppl::bexecute failed: %s", sts.status_name());
	}
    }
  else
    {
      sts = proc.bexecute(init_call, &data_cb, (GCBTagType) 0);
      if (sts.status() != ASC_success)
	{
	  terminate_application();  
	  cone_error_msg("Process::bexecute failed: %s", sts.status_name());
	}
    }
  cerr << "done.\n";
}

/****************************************************************************
* start application
****************************************************************************/

void start_application()
{
  cerr <<  "Starting application..." << endl;

  if ( ispoe )
    {
      sts = pappl.bstart();
      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  cone_error_msg("PoeAppl::bstart failed %s", sts.status_name());
	}
    }
  else
    {
      sts = proc.bstart();
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  cone_error_msg("Process::bstart failed %s", sts.status_name());
	}
    }
}

/****************************************************************************
* main
****************************************************************************/

int main(int argc, char* argv[])
{
  parse_cmdl(argc, argv);
  init_dpcl(); 
  create_application(argc, argv);

  cerr << "Parsing application...          ";
   
  exec = new Executable(proc, basename(prog_name), mainfun);
  cerr << "done.\n";
  
  cerr << "Building Call-Graph...          ";
  cgraph = new Callgraph(*exec, mainfun); 
  cerr << "done.\n"; 
  
  std::ofstream out(CONE_APPL_DESCR_FILE);

  out << *exec; 
  out << *cgraph;
  out.close();
  
  locate_probe_functions();
  instrument_functions();
  instrument_call_sites();
  init_probes();
  start_application();
  remove(CONE_APPL_DESCR_FILE);
  Ais_main_loop(); 
  cerr << "\nCONE  end.\n";
  return 0;
}


