/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "cone_cg.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "cone_metv.h"
#include "cone_cndata.h"
#include "cone_error.h"

#define CONE_MAX_STRING 4096

static struct cone_metv* metv = NULL ;
static struct cone_prog* cg_prog = NULL ; 
int ncnodes;
int rank = 0; 
int num_procs = 1;
struct cone_cnode* cone_cnode; 


struct timeval *tv;
struct timezone *tz;
  
void cone_cg_init(struct cone_prog* prog)
{
  cg_prog = prog;  
  metv = cone_metv_create();  
  ncnodes = cone_prog_num_cnodes(cg_prog); 
} 

void cone_cg_enter(int cnode_id)
{
  cone_cndata_enter(cone_cnode_get_data(cone_prog_cnode(cg_prog, cnode_id)), metv);
}


void cone_cg_exit(int cnode_id)
{
  cone_cndata_exit(cone_cnode_get_data(cone_prog_cnode(cg_prog, cnode_id)), metv);
}


void cone_cg_after_mpi_init(int my_rank, int num_ranks)
{ 
  rank = my_rank;
  num_procs = num_ranks;
}

int cone_cg_met_count()
{
  return cone_metv_size();
}


void cone_cg_print_cndata(FILE* fp) 
{ 
  int  ncnodes;
  int  i;
  char host_name[CONE_MAX_STRING];
  

  if ( gethostname(host_name, CONE_MAX_STRING) < 0 ) 
    cone_error(); 

  fprintf(fp, "HOST %s\n", host_name);

  ncnodes = cone_prog_num_cnodes(cg_prog);

  for ( i = 0 ; i< ncnodes ; i++ )
    cone_cndata_fprintf(fp, cone_cnode_get_data(cone_prog_cnode(cg_prog, i)));
 
  fclose(fp);
} 


void cone_cg_print_appdata(FILE* fp) 
{
  int i;   

  fprintf(fp, "PROGRAM %ld %ld %ld %ld %d %s\n", 
	  cone_prog_num_mods(cg_prog),  
	  cone_prog_num_funcs(cg_prog),  
	  cone_prog_num_csites(cg_prog),
	  cone_prog_num_cnodes(cg_prog),
	  cone_metv_size(metv) + cone_cndata_num_stdmets(),
	  cone_prog_name(cg_prog));

  for ( i = 0; i < cone_prog_num_mods(cg_prog); i++)
    cone_mod_fprintf(fp, cone_prog_mod(cg_prog, i));
  for ( i = 0; i < cone_prog_num_funcs(cg_prog); i++)
    cone_func_fprintf(fp, cone_prog_func(cg_prog, i));
  for ( i = 0; i < cone_prog_num_csites(cg_prog); i++)
    cone_csite_fprintf(fp, cone_prog_csite(cg_prog, i));
  for ( i = 0; i < cone_prog_num_cnodes(cg_prog); i++)
    cone_cnode_fprintf(fp, cone_prog_cnode(cg_prog, i));
  for ( i = 0; i < cone_cg_met_count(); i++)
    cone_metric_fprintf(fp, cone_metv_metric(i));
  
  cone_cndata_fprintf_stdmets(fp);
  
  fclose(fp);
}


void cone_cg_term()
{
  int   i;  
  FILE* f;
  FILE* fp;
  char  cndata_file[CONE_MAX_STRING];
  char  cmd[CONE_MAX_STRING];

  /* print the adf data in output.cdf */

   if ( rank == 0 )
    {
      if ( ( f = fopen("output.cdf", "w+")) == 0 ) 
	cone_error_msg("Error in opening the file output.cdf.\n");

      cone_cg_print_appdata(f); 
    }

  /* print the cnode data in output.rank.cdf*/

  sprintf(cndata_file, "output.%d.cdf", rank);
 
  if ( ( fp = fopen(cndata_file, "w+")) == 0 ) 
    cone_error_msg("Error in opening the file %s.\n", cndata_file);

  cone_cg_print_cndata(fp);
  fclose(fp);
  
  for( i = 0; i < ncnodes ; i++ )
    cone_cndata_delete(cone_cnode_get_data(cone_prog_cnode(cg_prog, i)));
 
  cone_metv_free(metv);

  cone_metv_close();
  
  /* merge the CONE per-process output files into a single CUBE file */

  if ( rank != 0 )
    return;

  /*- wait for files to be ready -*/
  for ( i = 0; i < num_procs; i++ )
    {
      int nf = 0;
      sprintf(cndata_file, "output.%d.cdf", i);
      cone_cntl_msg("Checking for %s ...", cndata_file);
    
      while (access(cndata_file, R_OK) != 0 )
	{
	  ++nf;
	  /*- if file not ready in 15 sec give up -*/
	  if ( nf > 15 ) 
	    cone_error_msg("merge failed");
	  sleep(1);
	}
    }

  sprintf(cmd, "cone_merge %d", num_procs);
  system(cmd);
}
