/****************************************************************************
**  Copyright (C) 2004                                                     **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "cone_error.h"

#include "Cube.h" 

#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <map>

#define CONE_MAX_STRING 4096

using namespace std;

/*
 *---------------------------------------------------------------------------------
 * data types 
 *---------------------------------------------------------------------------------
 */

struct metric
{
  int    id;
  string name;
  string descr;
  string uom;
  string mode;
  int    prop_id;
};

/*
 *---------------------------------------------------------------------------------
 * global variables 
 *---------------------------------------------------------------------------------
 */

// input-file prefix
char* prefix = "output";

// CONE id -> CUBE id 
map<int, int>       mod2mod;
map<int, int>       func2reg;
map<int, int>       csite2csite;
map<int, int>       cnode2csite;
map<int, int>       cnode2cnode;
vector<int>         root_cnodev;
map<int, metric>    id2met;
map<string, int>    name2metid;
map<int, int>       proc2thrd;
map<string, int>    host2node;

// call tree
map<int, vector<int> > cnode2children; 

// proc <-> host
map<int, string> proc2host;

// program name
string prog_name;

// machine id
int mach_id;

// (met, cnode) -> data
map<int, map<int, double> > mc2data;

// predefined property tree

map<string, string> prop2parent;

/*
 *---------------------------------------------------------------------------------
 * module 
 *---------------------------------------------------------------------------------
 */

void read_mod_data(ifstream& file, Cube& cube)
{
  string header, name, tmp;
  int id;

  file >> header >> id >> name;
  if ( header != "MODULE" )
    cone_error_msg("invalid MODULE record");
  
  mod2mod[id] = cube.def_module(name);
  getline(file, tmp);  
}

/*
 *---------------------------------------------------------------------------------
 * function 
 *---------------------------------------------------------------------------------
 */

void read_func_data(ifstream& file, Cube& cube)
{
  string header, name, tmp;
  int id, begln, endln, mod_id;

  file >> header >> id >> name >> begln >> endln >> mod_id;
  if ( header != "FUNCTION" )
    cone_error_msg("invalid FUNCTION record");
  
  func2reg[id] = cube.def_region(name, begln, endln, "", mod2mod[mod_id]);
  getline(file, tmp);  
}

/*
 *---------------------------------------------------------------------------------
 * call-site 
 *---------------------------------------------------------------------------------
 */

void read_csite_data(ifstream& file, Cube& cube)
{
  string header, tmp;
  int id, index, mod_id, line, caller_id, callee_id;
  
  file >> header >> id >> index >> mod_id >> line >> caller_id >> callee_id;  
  if ( header != "CALLSITE" )
    cone_error_msg("invalid CALLSITE record");

  csite2csite[id] = cube.def_csite(mod2mod[mod_id], line, func2reg[callee_id]);
  getline(file, tmp); 
}

/*
 *---------------------------------------------------------------------------------
 * cnode (i.e., call path)
 *---------------------------------------------------------------------------------
 */

void read_cnode_data(ifstream& file, Cube& cube)
{
  string header, tmp;
  int id, csite_id, parent_id, nchilds, child_id;

  file >> header >> id >> csite_id >> parent_id >> nchilds;  
  if ( header != "CNODE" )
    cone_error_msg("invalid CNODE record");

  for ( int i = 0; i < nchilds; i++)
    {
      file >> child_id;
      cnode2children[id].push_back(child_id);
    }
  
  if ( parent_id == -1)
    root_cnodev.push_back(id);

  cnode2csite[id]  = csite_id;
  getline(file, tmp);  
}

void create_call_tree(int cnode_id, int parent_id, Cube& cube)
{
  if ( parent_id == -1 )
      cnode2cnode[cnode_id] = cube.def_cnode(cnode2csite[cnode_id], -1);
  else
      cnode2cnode[cnode_id] = cube.def_cnode(cnode2csite[cnode_id], cnode2cnode[parent_id]);

  for ( int j = 0; j < cnode2children[cnode_id].size(); j++ )
    create_call_tree(cnode2children[cnode_id][j], cnode_id, cube);  
}

/*
 *---------------------------------------------------------------------------------
 * metric 
 *---------------------------------------------------------------------------------
 */

void read_met_data(ifstream& file, Cube& cube)
{
  string header;
  struct metric met;

  file >> header >> met.id >> met.name >> met.uom >> met.mode;
  getline(file, met.descr);

  if ( header != "METRIC" )
    cone_error_msg("invalid METRIC record");

  met.prop_id = -1;
  id2met[met.id]       = met;
  name2metid[met.name] = met.id;  
}

void predefine_prop_tree()
{  
  prop2parent["L1_D_MISS"]   = "L1_D_ACCESS";
  prop2parent["L1_D_HIT"]    = "L1_D_ACCESS";
  prop2parent["TLB_I_MISS"]  = "TLB_MISS";
  prop2parent["TLB_D_MISS"]  = "TLB_MISS";
  prop2parent["FP_INS"]      = "INSTRUCTION";
  prop2parent["FP_SQRT_INS"] = "FP_INS";
  prop2parent["FP_DIV_INS"]  = "FP_INS";  
}

void create_prop_tree(Cube& cube)
{
  bool found = false;

  for ( int i = 0; i < id2met.size(); i++ )
    {
      struct metric& met = id2met.find(i)->second;

      if ( met.prop_id == -1 )
	{
	  found = true;
	  
	  if ( prop2parent.find(met.name) != prop2parent.end() )
	    {
	      int parent_id  = name2metid[prop2parent[met.name]];
	      
	      if ( id2met[parent_id].prop_id != -1 )
		met.prop_id = cube.def_met(met.name, met.uom, met.descr, id2met[parent_id].prop_id);
	      else
		met.prop_id = cube.def_met(met.name, met.uom, met.descr, -1);
	    }
	  else
	    {
	      met.prop_id = cube.def_met(met.name, met.uom, met.descr, -1);
	    }
	}
    }
  if ( found )
    create_prop_tree(cube);
  else
    return;
    
}

/*
 *---------------------------------------------------------------------------------
 * read data referring to the entire program
 *---------------------------------------------------------------------------------
 */

void read_prog_data(Cube& cube)
{
  char file_name[CONE_MAX_STRING];
  string header;
  int nmods, nfuncs, ncsites, ncnodes, nmets;
    
  sprintf(file_name, "%s.cdf", prefix);

  ifstream file(file_name);
  if ( !file )
    cone_error_msg("file not found");
  
  file >> header >> nmods >> nfuncs >> ncsites >> ncnodes >> nmets >> prog_name;  
  if ( header != "PROGRAM" )
    cone_error_msg("invalid PROGRAM record");

  for ( int i = 0; i < nmods; i++)
      read_mod_data(file, cube);
  for ( int i = 0; i < nfuncs; i++)
      read_func_data(file, cube);
  for ( int i = 0; i < ncsites; i++)
      read_csite_data(file, cube);
  for ( int i = 0; i < ncnodes; i++)
      read_cnode_data(file, cube);
  for ( int i = 0; i < nmets; i++)
      read_met_data(file, cube);

  file.close();

  predefine_prop_tree();
  create_prop_tree(cube);

  for ( int i = 0; i < root_cnodev.size(); i++ )
    create_call_tree(root_cnodev[i], -1, cube);
}

/*
 *---------------------------------------------------------------------------------
 * host
 *---------------------------------------------------------------------------------
 */

void read_host_data(ifstream& file, Cube& cube, int proc_id)
{
  string header, name, tmp;

  file >> header >> name;  
  if ( header != "HOST" )
    cone_error_msg("invalid HOST record");

  proc2host[proc_id] = name;
  if ( host2node.find(name) == host2node.end() )
    host2node[name] = cube.def_node(name, mach_id);
  proc2thrd[proc_id] = cube.def_thrd("", cube.def_proc("", host2node[name]));
  getline(file, tmp);
}


/*
 *---------------------------------------------------------------------------------
 * call-tree node data
 *---------------------------------------------------------------------------------
 */

void read_cndata_data(ifstream& file, Cube& cube, int proc_id)
{
  string header, tmp;
  int cnode_id;
  unsigned long long uvalue;
  double dvalue;

  file >> header >> cnode_id;  
  if ( header != "CNDATA" )
    cone_error_msg("invalid CNDATA record");

  for ( int i = 0; i < id2met.size(); i++ )
    if ( id2met[i].uom == "sec" )
      {
	file >> dvalue;
	mc2data[i][cnode_id] = dvalue;
      }
    else
      {
	file >> uvalue;
	mc2data[i][cnode_id] = uvalue;
      }
  getline(file, tmp);
}

/*
 *---------------------------------------------------------------------------------
 * read data referring to a single process
 *---------------------------------------------------------------------------------
 */

void read_proc_data(Cube& cube, int proc_id)
{
  char file_name[CONE_MAX_STRING];

  sprintf(file_name, "%s.%d.cdf", prefix, proc_id);

  ifstream file(file_name);

  if ( !file )
    cone_error_msg("file not found");

  read_host_data(file, cube, proc_id);
  for ( int i = 0; i < cnode2cnode.size(); i++)
    read_cndata_data(file, cube, proc_id); 

  // transform inclusive values into exclusive values
  for ( int i = 0; i < id2met.size(); i++)
      for ( int j = 0; j < cnode2cnode.size(); j++)
	{
	  cube.set_sev(id2met[i].prop_id, cnode2cnode[j], proc2thrd[proc_id], mc2data[i][j]);
	  if ( id2met[i].mode == "incl" )
	    for ( int k = 0; k < cnode2children[j].size(); k++)
	      cube.sub_sev(id2met[i].prop_id, 
			   //			   cnode2cnode[cnode2children[j][k]], 
			   cnode2cnode[j], 
			   proc2thrd[proc_id], 
			   mc2data[i][cnode2children[j][k]]);
	}
	
  file.close();
}

/*
 *---------------------------------------------------------------------------------
 * main
 *---------------------------------------------------------------------------------
 */
  
int main(int argc, char** argv)
{
  char* rest;
  int infilec = 0;
  char file_name[CONE_MAX_STRING];

  /* ckeck # of arguments */
  if (argc < 2 || argc > 3)
    cone_error_msg("Wrong # of arguments\nUsage: cone_merge <#files> "
	          "[<prefix>]");
    
  infilec = strtol(argv[1], &rest, 10);
  if (rest == argv[1])
    cone_error_msg("Integer expected as 1st argument\nUsage: cone_merge "
	          "<#files> [<prefix>]");

  if (argc > 2)
    prefix = argv[2];

 
  /* merge */
  Cube cube;

  /* program data */
  read_prog_data(cube);

  
  /* location hierarchy */
  mach_id = cube.def_mach("IBM", cube.def_grid("Grid"));

  // printf("reading per-process data\n");

  /* per-process data */
  for ( int i = 0; i < infilec; i++ )
    read_proc_data(cube, i);
    

  /* write CUBE file */  
  char cube_file_name[CONE_MAX_STRING];
  sprintf(cube_file_name, "%s.cube", prog_name.c_str());
  ofstream cube_file(cube_file_name);
  if ( !cube_file )
    cone_error_msg("cannot open file %s", cube_file_name);
  
  cube_file << cube;
  cube_file.close();
  cone_cntl_msg("Local output files merged");
  
  /* remove per-process output files */
  for ( int i = 0; i < infilec; i++ )
    {
      sprintf(file_name, "%s.%d.cdf", prefix, i);

      int result = remove(file_name);  
      if (result == 0)
	cone_cntl_msg("Removed output file %s", file_name);
      else
	cone_error_msg("Cannot remove output file %s", file_name);
    } 

  /* remove output.cdf */
  sprintf(file_name, "%s.cdf", prefix);
  int result = remove(file_name);  
  if (result == 0)
    cone_cntl_msg("Removed output file %s", file_name);
  else
    cone_error_msg("Cannot remove output file %s", file_name); 

  exit(0);
}
















